==> xcoral_3.40.orig.tar.gz: xcoral-3.40/Smac/funcall.c <==
/* ########################################################################

			       funcall.c

   File: funcall.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/funcall.c
   Description: 
   Created: Tue Feb 21 12:53:55 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 12:53:56 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



#include <stdio.h>

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include "funcall.h"
#include "FunctionCall.h"
#include "Builtin.h"
#include "list.h"
#include "hash.h"
#include "error.h"

/*
  fct = make_funcall (fonction non calculee)
  info = (char **) [nbrearg, fct, arg1 ..] avec argi,fct : List *, nbrearg >= 0
*/

int bad_number_of_args(fct, nbrearg)
     Function * fct;
     int nbrearg;
{
  if (Function__NbreArg(fct) != nbrearg) {
    if (! Function__IsBuiltin(fct))
      return 1;
    else {
      int min = (int) HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct);

      if ((! min) || (nbrearg < (min - 1)))
	return 1;
    }
  }
  
  return 0;
}

Instruction * make_funcall(l)
     List * l;
{
  List ** largs = (List **) l->info;
  int nbrearg = (int) *largs++;
  Instruction * result;

  /* Ici on est ANSI : si la fonction n'est pas calculee ou est une variable
     globale il y a un appel a make_identifier qui DOIT trouver l'identifier */

  Function * fct = (Function *) MAKE(*largs);

  largs += 1;

  if (IsaFunction((Instruction *) fct)) {
    
    /*******************************
      La fonction n'est pas calculee
      ******************************/
    
    if (bad_number_of_args(fct, nbrearg)) {
      sprintf(err_msg, "%d arguments for %s, %s%d required",
	      nbrearg, Function__Name(fct),
	      (HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct))
	      ? "at least " : "",
	      Function__NbreArg(fct));
      Error(err_msg);
      result = 0;	/* to not have warning */
    }
    else {
      int nargtheo = Function__NbreArg(fct);
      int index;
      Instruction ** args;
      
#if defined(RUNTIMECHECK) || defined(XCORAL)
      if (HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct)) {
	args = (Instruction **) Malloc((nbrearg + 1) * sizeof(Instruction *));
	args[nbrearg] =
	  (Instruction *) NotEnoughArg__NotEnoughArg(fct, nbrearg);
      }
      else
	args = (Instruction **) Malloc(nbrearg * sizeof(Instruction *));
#else
      args = (Instruction **) Malloc(nbrearg * sizeof(Instruction *));
#endif
    
      for (index = 0; index != nbrearg; index += 1) {
	args[index] = MAKE(*largs);
	largs += 1;
	if (nargtheo) {
	  nargtheo -= 1;
	  if (! ExprType__Convertible(args[index],
				      Function__ArgType(fct, index))) {
	    sprintf(err_msg, "incompatible type for %s argument %d, ",
		    Function__Name(fct), index + 1);
	    Type__Print(Function__ArgType(fct, index), err_msg);
	    strcat(err_msg, " <> ");
	    Type__Print(GetExprType(args[index]), err_msg);
	    Error(err_msg);
	  }
	}
      }

      if (Function__IsBuiltin(fct))
	result = (Instruction *)
	  BuiltinFunctionCall__BuiltinFunctionCall(fct, args, nbrearg,
						   Function__ValType(fct));
      else
	result = (Instruction *)
	    FunctionCall__FunctionCall(fct, args, nbrearg);
    }
  }
  else if (Type__IsaFunction(GetExprType((Instruction *) fct))) {
    TypeFunction * tfct = Type__Function(GetExprType((Instruction *) fct));
    
    /*************************
      La fonction est calculee
      ************************/

    if (TypeFunction__NbreArg(tfct) > nbrearg) {
      sprintf(err_msg, "%d arguments at least %d required",
	      nbrearg, TypeFunction__NbreArg(tfct));
      Error(err_msg);
      result = 0;	/* to not have warning */
    }
    else {
      int index;
      Instruction ** args =
	(Instruction **) Malloc(nbrearg * sizeof(Instruction *));
    
      for (index = 0; index != nbrearg; index += 1) {
	args[index] = MAKE(*largs);
	largs += 1;
	if (! ExprType__Convertible(args[index],
				    TypeFunction__ArgType(tfct, index))) {
	  sprintf(err_msg, "incompatible type argument %d, ", index + 1);
	  Type__Print(TypeFunction__ArgType(tfct, index), err_msg);
	  strcat(err_msg, " <> ");
	  Type__Print(GetExprType(args[index]), err_msg);
	  Error(err_msg);
	}
      }

      result = (Instruction *)
	CalculateFunctionCall__CalculateFunctionCall((Instruction *) fct,
						       args, nbrearg);
    }
  }
  else {
    Error("illegal function call");
    result = 0;	/* to not have warning */
  }
  
  free(l->info);
  free(l);
  
  return result;
}

==> xcoral-3.44.tar.gz: xcoral-3.44/Smac/funcall.c <==
/* ########################################################################

			       funcall.c

   File: funcall.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/funcall.c
   Description: 
   Created: Tue Feb 21 12:53:55 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 12:53:56 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



#include <stdio.h>

#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include "funcall.h"
#include "FunctionCall.h"
#include "Builtin.h"
#include "list.h"
#include "hash.h"
#include "error.h"

#include "decl.h" /**/

/*
  fct = make_funcall (fonction non calculee)
  info = (char **) [nbrearg, fct, arg1 ..] avec argi,fct : List *, nbrearg >= 0
*/

int bad_number_of_args(fct, nbrearg)
     Function * fct;
     int nbrearg;
{
  if (Function__NbreArg(fct) != nbrearg) {
    if (! Function__IsBuiltin(fct))
      return 1;
    else {
      int min = (int) HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct);

      if ((! min) || (nbrearg < (min - 1)))
	return 1;
    }
  }
  
  return 0;
}

static void funcall_error(l, largs, index, nbrearg, args)
    List * l;
    List ** largs;
    int index;
    int nbrearg;
    Instruction ** args;
{
  if (args != 0) {
    int i;
    
    for (i = 0; i != index; i += 1)
      /* Delete the already made instructions */
      Delete(args[i]);

    index += 1;	/* skip the erroned case */
    
    free(args);
  }
  
  while (++index < nbrearg) {
    /* delete the not already made instruction */
    DELETE(*largs);
    largs += 1;
  }
  
  free(l->info);
  free(l);
}

Instruction * make_funcall(l, del)
    List * l;
    int del;
{
  List ** largs = (List **) l->info;
  int nbrearg = (int) *largs++;
  Instruction * result;
  Function * fct;
  int index;

  if (del) {
    funcall_error(l, largs, 0, nbrearg + 1, 0);
    return 0;
  }
  
  /* Ici on est ANSI : si la fonction n'est pas calculee ou est une variable
     globale il y a un appel a make_identifier qui DOIT trouver l'identifier */

  fct = (Function *) MAKE(*largs);

  largs += 1;
  
  if (((Instruction *) fct) == ERROR) {
    funcall_error(l, largs, 0, nbrearg, 0);
    return ERROR;
  }
  else if (IsaFunction((Instruction *) fct)) {
    
    /*******************************
      La fonction n'est pas calculee
      ******************************/
    
    if (bad_number_of_args(fct, nbrearg)) {
      sprintf(err_msg, "%d arguments for %s, %s%d required",
	      nbrearg, Function__Name(fct),
	      (HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct))
	      ? "at least " : "",
	      Function__NbreArg(fct));
      Error(err_msg, 1);
      funcall_error(l, largs, 0, nbrearg, 0);
      return ERROR;
    }
    else {
      int nargtheo = Function__NbreArg(fct);
      Instruction ** args;
      
#if defined(RUNTIMECHECK) || defined(XCORAL)
      int have_notenougth;
      
      if (HashTable__Search(builtin_fct_nbre_arg_min, (Object) fct)) {
	have_notenougth = 1;
	args = (Instruction **) Malloc((nbrearg + 1) * sizeof(Instruction *));
	args[nbrearg] =
	  (Instruction *) NotEnoughArg__NotEnoughArg(fct, nbrearg);
      }
      else {
	have_notenougth = 0;
	args = (Instruction **)
	  Malloc((nbrearg) ? nbrearg * sizeof(Instruction *) : 1);
      }
#else
      args = (Instruction **)
	Malloc((nbrearg) ? nbrearg * sizeof(Instruction *) : 1);
#endif
    
      for (index = 0; index != nbrearg; index += 1) {
	args[index] = MAKE(*largs);
	largs += 1;
	if (args[index] == ERROR) {
#if defined(RUNTIMECHECK) || defined(XCORAL)
	  if (have_notenougth)
	    Delete(args[nbrearg]);
#endif
	  funcall_error(l, largs, index, nbrearg, args);
	  return ERROR;
	}
	else if (nargtheo) {
	  ConvFct conv =
	    ExprType__Convertible(args[index], Function__ArgType(fct, index));
	      
	  nargtheo -= 1;
	  if (conv == 0) {
	    sprintf(err_msg, "incompatible type for %s argument %d, ",
		    Function__Name(fct), index + 1);
	    Type__Print(Function__ArgType(fct, index), err_msg);
	    strcat(err_msg, " <> ");
	    Type__Print(GetExprType(args[index]), err_msg);
	    Error(err_msg, 1);
	    Delete(args[index]);
#if defined(RUNTIMECHECK) || defined(XCORAL)
	    if (have_notenougth)
	      Delete(args[nbrearg]);
#endif
	    funcall_error(l, largs, index, nbrearg, args);
	    return ERROR;
	  }
	  args[index] = Conv__Conv(args[index], conv);
	}
      }

      if (Function__IsBuiltin(fct))
	result = (Instruction *)
	  BuiltinFunctionCall__BuiltinFunctionCall(fct, args, nbrearg,
						   Function__ValType(fct));
      else
	result = (Instruction *)
	    FunctionCall__FunctionCall(fct, args, nbrearg);
    }
  }
  else if (Type__IsaFunction(GetExprType((Instruction *) fct))) {
    TypeFunction * tfct = Type__Function(GetExprType((Instruction *) fct));
    
    /*************************
      La fonction est calculee
      ************************/

    if (TypeFunction__NbreArg(tfct) > nbrearg) {
      sprintf(err_msg, "%d arguments at least %d required",
	      nbrearg, TypeFunction__NbreArg(tfct));
      Error(err_msg, 1);
      Delete((Instruction *) fct);
      funcall_error(l, largs, 0, nbrearg, 0);
      return ERROR;
    }
    else {
      int index;
      Instruction ** args =
	(Instruction **) Malloc((nbrearg) ? nbrearg * sizeof(Instruction *) : 1);
    
      for (index = 0; index != nbrearg; index += 1) {
	ConvFct conv;
	
	args[index] = MAKE(*largs);
	if (args[index] == ERROR) {
	  Delete((Instruction *) fct);
	  funcall_error(l, largs, index, nbrearg, args);
	  return ERROR;
	}
	largs += 1;
	conv = ExprType__Convertible(args[index],
				     TypeFunction__ArgType(tfct, index));
	if (conv == 0) {
	  sprintf(err_msg, "incompatible type argument %d, ", index + 1);
	  Type__Print(TypeFunction__ArgType(tfct, index), err_msg);
	  strcat(err_msg, " <> ");
	  Type__Print(GetExprType(args[index]), err_msg);
	  Error(err_msg, 1);
	  Delete((Instruction *) fct);
	  Delete(args[index]);
	  funcall_error(l, largs, index, nbrearg, args);
	  return ERROR;
	}
	else
	  args[index] = Conv__Conv(args[index], conv);
      }

      result = (Instruction *)
	CalculateFunctionCall__CalculateFunctionCall((Instruction *) fct,
						     args, nbrearg);
    }
  }
  else {
    Error("illegal function call", 1);
    Delete((Instruction *) fct);
    funcall_error(l, largs, 0, nbrearg, 0);
    return ERROR;
  }
  
  free(l->info);
  free(l);
  
  return result;
}
